import json
import copy
import traceback

from iot_devices.host import get_class, discover
from iot_devices.device import Device

print("# Known Device Plugins\n")
print("All devices shown, some config params may be unlisted.")
print("This documentation is autogenerated with iot_devices_scan.py\n")
d = discover()
for i in sorted(d.keys()):
    print("## " + i+"\n")
    print(d[i].get('description', '')+'\n')
    try:
        c1 = get_class({'type': i})

        dpi = {}
        dpd = {}
        dpw = {}

        props = {}

        class c(c1):
            def string_data_point(self,
                                  name: str,
                                  description: str = "",
                                  unit: str = '',
                                  handler=None,
                                  interval: float = 0,
                                  writable=True,
                                  subtype: str = '',
                                  **kwargs):
                Device.string_data_point(self, name, description=description, unit=unit,
                                         handler=handler, interval=interval, writable=writable,
                                         subtype=subtype, **kwargs)
                dpi[name] = f"{subtype} {unit}"
                dpd[name] = description
                dpw[name] = writable

            def numeric_data_point(self,
                                   name: str,
                                   description: str = "",
                                   unit: str = '',
                                   handler=None,
                                   interval: float = 0,
                                   writable=True,
                                   subtype: str = '',
                                   **kwargs):
                Device.numeric_data_point(self, name, description=description, unit=unit,
                                          handler=handler, interval=interval, writable=writable,
                                          subtype=subtype, **kwargs)
                dpi[name] = f"""{subtype} {unit}""".strip()
                dpd[name] = description
                dpw[name] = writable

        defaults = {}
        tags = {}

        try:
            inst = c('DeviceName', {'type': i})
            props.update(inst.config_properties)

            for j in inst.config:
                if not j in props:
                    props[j] = {}

            tags.update(inst.datapoints)

        except Exception:
            print(traceback.format_exc())

        print("```python")
        print("from iot_devices.host import create_device")
        print("from " + d[i]['importable'] + " import " + i + "\n")

        print("dev = " + 'create_device(' + i + ', "name", {')
        count = 0
        for j in props:
            count += 1
            if j.startswith("device."):
                desc = props[j].get('description', '')
                if desc:
                    print(
                        '    # '+props[j].get('description', '')+"")
                ending = "," if count < len(props) else '\n})'
                if j in inst.config:
                    print(f"    '{j}': '{inst.config[j]}'"+ending)

        print("\n\n")
        for j in dpi:
            desc = dpd.get(j, '')
            if dpi[j].strip():
                print(f'# {dpi[j]}')
            if desc:
                print(f"# {desc}")
            print(f"print(dev.datapoints['{j}'])")
            print(f"# >>> {str(inst.datapoints[j])[:64]}\n")

            if dpw.get(j, False):
                print(f"# {j} is writable")
                print(
                    f"# dev.set_data_point('{j}', <your value> )\n")
        print("```")

    except Exception:
        print("## Error getting data for " + i)
        print(traceback.format_exc())

    print("")



import json
import iot_devices.host as host
import iot_devices.device as devices
from scullery import mqtt
import logging
import time
import threading
import os
import weakref

lock = threading.Lock()


all_devs = weakref.WeakValueDictionary()


mqttlock = threading.Lock()

stopFlag = [0]


def scan():
    while 1:
        if stopFlag[0]:
            break
        time.sleep(10)
        if lock.acquire(timeout=1):
            try:
                for i in all_devs:
                    # If the last signal was very strong, we don't need to wait as long before considering
                    # it gone, because packet loss will be less
                    m = 3 if (all_devs[i].datapoints["rssi"] or -80) > -65 else 7

                    if all_devs[i].lastseen < time.time() - (
                        float(all_devs[i].config.get("interval", 300) or 300) * m
                    ):
                        # This is how we mark it as not there
                        all_devs[i].set_data_point("rssi", -180)
            except Exception:
                logging.exception("RTL err")

            finally:
                lock.release()


t = threading.Thread(target=scan, name="RTL433Task", daemon=True)
t.start()


def stop():
    stopFlag[0] = 1


host.app_exit_register(stop)


defaultSubclassCode = """
class CustomDeviceType(DeviceType):
    pass
"""


class RTL433Client(devices.Device):
    device_type = "RTL433Client"
    readme = os.path.join(os.path.dirname(__file__), "README.md")
    defaultSubclassCode = defaultSubclassCode
    shortDescription = (
        "This device lets you get data from a device using an RTL433 daemon and MQTT"
    )

    def onConnectionChange(self, status):
        if status == "connected":
            self.set_data_point("mqttStatus", 1)
        else:
            self.set_data_point("mqttStatus", 0)

    def __init__(self, name, data, **kw):
        devices.Device.__init__(self, name, data, **kw)

        try:
            self.set_config_default("device.interval", "300")

            self.numeric_data_point(
                "rssi",
                min=-180,
                max=12,
                interval=float(self.config["device.interval"]),
                description="-75 if recetly seen, otherwise -180, we don't have real RSSI data",
                writable=False,
            )

            self.set_alarm("No Signal", "rssi", "value<-110")

            self.set_config_default("device.id", "")
            self.set_config_default("device.model", "")
            self.set_config_default("device.server", "localhost")
            self.set_config_default("device.port", "1883")
            self.set_config_default("device.password", "")
            self.set_config_default("device.mqttTopic", "home/rtl_433")

            # Pretend we have seen it to give the checker the right interval
            # before declaring it lost
            self.lastseen = time.time()

            self.connection = mqtt.get_connection(
                self.config["device.server"],
                int(self.config["device.port"].strip() or 1883),
                password=self.config["device.password"].strip(),
                connection_id="RTL433Connection",
            )

            self.numeric_data_point("mqttStatus", writable=False)
            self.connection.subscribe_to_status(self.onConnectionChange)
            self.set_data_point("mqttStatus", 1 if self.connection.is_connected else 0)
            self.set_alarm(
                "MQTT Lost", "mqttStatus", "value < 0.5", auto_ack=True, trip_delay=10
            )

            topic = data.get("device.mqtttopic", "home/rtl_433")

            # We cannot use priority greater than info because these are unencrypted and untruusted and higher could make noise.

            def onBattery(t, m):
                m = float(m)
                if "battery" not in self.datapoints:
                    self.numeric_data_point(
                        "battery", default=50, writable=False, unit="%"
                    )

                    # Always set before setting the alarm.
                    self.set_data_point("battery", m)

                    self.set_alarm(
                        name="Low battery",
                        datapoint="battery",
                        expression="value < 15",
                        priority="info",
                        auto_ack=True,
                    )

                self.set_data_point("battery", m)

            def onWind(t, m):
                m = float(m)
                if "wind" not in self.datapoints:
                    self.numeric_data_point("wind", unit="km/h", writable=False)

                    self.set_alarm(
                        name="High Wind",
                        datapoint="wind",
                        expression="value > 35",
                        priority="info",
                    )

                self.set_data_point("wind", m)

            def onTemp(t, m):
                m = float(m)
                if "temp" not in self.datapoints:
                    self.numeric_data_point("temp", unit="degC", writable=False)
                    self.set_alarm(
                        name="Freezing temperatures",
                        datapoint="temp",
                        expression="value < 2",
                        priority="info",
                    )

                self.set_data_point("temp", m)

            def onHum(t, m):
                m = float(m)
                if "humidity" not in self.datapoints:
                    self.numeric_data_point("humidity", unit="%", writable=False)
                    self.set_alarm(
                        name="High humidity",
                        datapoint="humidity",
                        expression="value > 80",
                        priority="info",
                    )
                    self.set_alarm(
                        name="Low humidity",
                        datapoint="humidity",
                        expression="value < 20",
                        priority="info",
                        auto_ack=True,
                    )

                self.set_data_point("humidity", m)

            def onMoist(t, m):
                m = float(m)
                if "moisture" not in self.datapoints:
                    self.numeric_data_point("moisture", unit="%", writable=False)
                self.set_data_point("moisture", m)

            def onPres(t, m):
                m = float(m)
                if "pressure" not in self.datapoints:
                    self.numeric_data_point("pressure", unit="Pa", writable=False)
                self.set_data_point("pressure", m)

            def onWeight(t, m):
                m = float(m)
                if "weight" not in self.datapoints:
                    self.numeric_data_point("weight", writable=False)
                self.set_data_point("weight", m)

            def onCommandCode(t, m):
                m = float(m)
                if "lastCommandCode" not in self.datapoints:
                    self.object_data_point("lastCommandCode", writable=False)
                self.set_data_point("lastCommandCode", (m, time.time()))

            def onCommandName(t, m):
                m = float(m)
                if "lastCommandName" not in self.datapoints:
                    self.object_data_point("lastCommandName", writable=False)
                self.set_data_point("lastCommandName", (m, time.time()))

            def onJSON(t, m):
                try:
                    m = json.loads(m)
                    self.print(m, "Saw packet on air")

                    # Going to do an ID match.
                    if "device.id" in self.config and self.config["device.id"]:
                        if not ("id" in m and str(m["id"]) == self.config["device.id"]):
                            self.print(m, "Packet filter miss")
                            return

                    if "device.model" in self.config and self.config["device.id"]:
                        if not (
                            "model" in m and m["model"] == self.config["device.model"]
                        ):
                            self.print(m, "Packet filter miss")
                            return

                    self.print(m, "Packet filter hit")

                    # No real RSSI, we just use -75 to mean good and -180 to mean bad.
                    self.set_data_point("rssi", -75)
                    self.set_alarm("Signal Lost", "rssi", "value < -98", auto_ack=True)

                    self.lastseen = time.time()

                    if "humidity" in m:
                        onHum(0, m["humidity"])

                    if "moisture" in m:
                        onMoist(0, m["moisture"])

                    if "temperature_C" in m:
                        onTemp(0, m["temperature_C"])

                    if "wind_avg_km_h" in m:
                        onWind(0, m["wind_avg_km_h"])

                    if "pressure_kPa" in m:
                        onPres(0, m["pressure_kPa"] * 1000)

                    if "pressure_hPa" in m:
                        onPres(0, m["pressure_hPa"] * 100)

                    # Keep a percent based API with randomly chosen high and low numbers
                    if "battery_ok" in m:
                        onBattery(0, 100 if m["battery_ok"] else 5)

                    if "cmd" in m:
                        onCommandCode(0, m["cmd"])

                    if "button_id" in m:
                        onCommandCode(0, m["button_id"])

                    if "button_name" in m:
                        onCommandName(0, m["button_name"])

                    if "event" in m:
                        onCommandName(0, m["event"])

                    if "code" in m:
                        onCommandCode(0, m["code"])
                except Exception:
                    self.handle_exception()

            self.noGarbage = [onJSON]

            self.connection.subscribe(topic, onJSON, encoding="raw")

            all_devs[self.name] = self
        except Exception:
            self.handle_exception()

    def close(self):
        return super().close()
        self.connection.unsubscribe(self.noGarbage[0])



import os.path
import os
import asyncio
from contextlib import closing
import socket
import time
from re import L
import uuid
from iot_devices import device
import threading
import logging


"""
httpserver code:
Copyright 2021 Brad Montgomery
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial
portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import argparse
from http.server import HTTPServer, BaseHTTPRequestHandler


x = """<?xml version="1.0" encoding="UTF-8" ?>
<root xmlns="urn:schemas-upnp-org:device-1-0">
<specVersion>
    <major>1</major>
    <minor>0</minor>
</specVersion>
<device>
    <deviceType>urn:roku-com:device:player:1-0</deviceType>
    <friendlyName>FN</friendlyName>
    <manufacturer>naimo84</manufacturer>
    <manufacturerURL>https://github.com/naimo84/</manufacturerURL>
    <modelDescription>Node Red - fake Roku player</modelDescription>
    <modelName>fakeroku</modelName>
    <modelNumber>4200X</modelNumber>
    <modelURL>https://github.com/naimo84/node-red-contrib-fakeroku</modelURL>
    <serialNumber>XXXX</serialNumber>
    <UDN>uuid:roku:ecp:XXXX</UDN>
    <iconList>
        <icon>
            <mimetype>image/png</mimetype>
            <width>360</width>
            <height>219</height>
            <depth>8</depth>
            <url>device-image.png</url>
        </icon>
    </iconList>
    <serviceList>
    <service>
        <serviceType>urn:roku-com:service:ecp:1</serviceType>
        <serviceId>urn:roku-com:serviceId:ecp1-0</serviceId>
        <controlURL/>
        <eventSubURL/>
        <SCPDURL>ecp_SCPD.xml</SCPDURL>
    </service>
    </serviceList>
</device>
</root>"""


def ssdpxml(name, uuid):
    return x.replace("FN", name).replace("XXXX", uuid.replace("-", "").upper())


logger = logging.Logger("plugins.pyremote")


# Attr: https://stackoverflow.com/questions/1365265/on-localhost-how-do-i-pick-a-free-port-number


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("0.0.0.0", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def check_port(p):
    try:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("0.0.0.0", p))
            return True
    except Exception:
        return False


try:
    import network
except Exception:
    pass

# https://github.com/FRC4564/HueBridge/blob/master/hue.py used as reference
# Note: You do not have to use SSDP's schemas. Look at roku, they do their own fields: https://developer.roku.com/docs/developer-program/debugging/external-control-api.md


try:
    ms = time.ticks_ms
    tickdiff = time.ticks_diff
except Exception:

    def ms():
        return int(time.time() * 1000)

    def tickdiff(a, b):
        return a - b


mxsearch = {
    "method": "M-SEARCH",
    "HOST": "239.255.255.250:1900",
    "MAN": '"ssdp:discover"',
    "MX": "3",
}

reply = {
    "SERVER": "Unspecified, UPnP/1.0, Unspecified",
    "EXT": "",
    "CACHE-CONTROL": "max-age=601",
}


# Local service definition:
def getUID():
    try:
        import network
        import ubinascii

        mac = ubinascii.hexlify(network.WLAN().config("mac"), ":").decode()
        return mac.replace(":", "").upper()
    except Exception:
        return socket.gethostname()


def getLocalIPForRemoteClient(addr):
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect(addr)
        ip = sock.getsockname()[0]
        sock.close()
    except Exception:
        ip = network.WLAN().ifconfig()[0]

    return ip


class HTTPUServer:
    def __init__(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(2)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        addr = ("", 1900)
        sock.bind(addr)
        opt = bytes([239, 255, 255, 250]) + bytes([0, 0, 0, 0])
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, opt)
        self.sock = sock

        self.services = {}

    def poll(self):
        try:
            data, addr = self.sock.recvfrom(1024)
        except OSError:
            return
        self._onmsg(addr, parse_httpu(data))

    def _replyServices(self, st, a):
        for i in self.services:
            if st == "ssdp:all" or st == i:
                s = {}
                s.update(self.services[i])
                s.update(reply)
                s["ST"] = s.get("ST", i)

                s["Location"] = s["Location"].replace(
                    "localhost", getLocalIPForRemoteClient(a)
                )

                s["LOCATION"] = s["LOCATION"].replace(
                    "localhost", getLocalIPForRemoteClient(a)
                )
                if not "USN" in s:
                    s["USN"] = "uuid:" + getUID() + "::" + s["ST"]
                s = make_httpu(s)
                self.sock.sendto(s, a)

    def _onmsg(self, a, m):
        if a[0].startswith("192.") or a[0].startswith("10.") or a[0].startswith("127."):
            if m.get("method", "") == "M-SEARCH":
                if m.get("Man", m.get("MAN", "")) == '"ssdp:discover"':
                    st = m.get("ST", "")
                    self._replyServices(st, a)


def make_httpu(d):
    o = ""
    if "method" in d:
        o += d["method"] + " * HTTP/1.1\r\n"
    else:
        o += "HTTP/1.1 200 OK\r\n"

    for i in d:
        if not i == "method":
            o += i + ": " + d[i] + "\r\n"
    o += "\r\n"

    return o.encode()


def parse_httpu(data):
    data = data.decode()

    d = {}
    # Compensate for any bad implementations that don't use \r\n
    lines = data.replace("\r", "").split("\n")

    l0 = lines.pop(0)

    if "*" in l0:
        d["method"] = l0.split("*")[0].strip()

    for l in lines:
        tokens = l.split(":", 1)
        if len(tokens) > 1:
            d[tokens[0]] = tokens[1].strip()
    return d


def tzget():
    tzname = os.environ.get("TZ")
    if tzname:
        return tzname
    elif os.path.exists("/etc/timezone"):
        with open("/etc/timezone") as f:
            return f.read().strip()
    return "US/Pacific"


def fakeroku(name):
    return (
        """<device-info>
    <serial-number>X004000B231</serial-number>
    <device-id>S00820BB231</device-id>
    <vendor-name>Roku</vendor-name>
    <model-name>Roku Ninja</model-name>
    <model-number>3930X</model-number>
    <model-region>US</model-region>
    <is-tv>false</is-tv>
    <is-stick>false</is-stick>
    <ui-resolution>1080p</ui-resolution>
    <supports-ethernet>false</supports-ethernet>
    <wifi-mac>10:59:32</wifi-mac>
    <wifi-driver>realtek</wifi-driver>
    <has-wifi-extender>false</has-wifi-extender>
    <has-wifi-5G-support>true</has-wifi-5G-support>
    <can-use-wifi-extender>true</can-use-wifi-extender>
    <network-type>wifi</network-type>
    <network-name>Hillary's Email Server</network-name>
    <friendly-device-name>Roku LivingRoom</friendly-device-name>
    <friendly-model-name>Roku Express</friendly-model-name>
    <default-device-name>Roku Express - X004000AJDX1</default-device-name>
    <user-device-name>"""
        + name
        + """</user-device-name>
    <user-device-location>LivingRoom</user-device-location>
    <build-number>AEA.00E04209A</build-number>
    <software-version>10.0.0</software-version>
    <software-build>4209</software-build>
    <secure-device>true</secure-device>
    <language>en</language>
    <country>US</country>
    <locale>en_US</locale>
    <time-zone-auto>true</time-zone-auto>
    <time-zone>"""
        + tzget()
        + """</time-zone>
    <time-zone-name>"""
        + tzget()
        + """</time-zone>
    <time-zone-tz>"""
        + tzget()
        + """</time-zone>
    <time-zone-offset>"""
        + str(time.localtime().tm_gmtoff / 60)
        + """</time-zone-offset>
    <clock-format>12-hour</clock-format>
    <uptime>2912968</uptime>
    <power-mode>PowerOn</power-mode>
    <supports-suspend>false</supports-suspend>
    <supports-find-remote>true</supports-find-remote>
    <find-remote-is-possible>false</find-remote-is-possible>
    <supports-audio-guide>true</supports-audio-guide>
    <supports-rva>true</supports-rva>
    <developer-enabled>false</developer-enabled>
    <keyed-developer-id/>
    <search-enabled>true</search-enabled>
    <search-channels-enabled>true</search-channels-enabled>
    <voice-search-enabled>true</voice-search-enabled>
    <notifications-enabled>true</notifications-enabled>
    <notifications-first-use>true</notifications-first-use>
    <supports-private-listening>true</supports-private-listening>
    <headphones-connected>false</headphones-connected>
    <supports-ecs-textedit>true</supports-ecs-textedit>
    <supports-ecs-microphone>true</supports-ecs-microphone>
    <supports-wake-on-wlan>false</supports-wake-on-wlan>
    <supports-airplay>true</supports-airplay>
    <has-play-on-roku>true</has-play-on-roku>
    <has-mobile-screensaver>false</has-mobile-screensaver>
    <support-url>roku.com/support</support-url>
    <grandcentral-version>5.5.62</grandcentral-version>
    <trc-version>3.0</trc-version>
    <trc-channel-version>4.2.3</trc-channel-version>
    <davinci-version>2.8.20</davinci-version>
    </device-info>"""
    )


class RokuRemoteApp(device.Device):
    device_type = "RokuRemoteApp"
    readme = """
Implements an extended version of the (https://developer.roku.com/docs/developer-program/debugging/external-control-api.md)[Roku ECP protocol].  Does not currently work with most real Roku apps,
intended mostly for use with DIY handheld remotes.

We condense everything down to a single "Command" tag.  "Launch" commands are mapped to a string like "launch:78797".
Keydown commands are mapped to a string containing the exact key name.

The battery tag represents the most recently connected remote that decided to send something. It should not be given all that much weight.
    """

    def ssdploop(self):
        while 1:
            s = self.ssdp
            if s:
                s.poll()
            else:
                return

    def close(self):
        self.ssdp = None
        self.bind = None
        if self.httpd:
            self.httpd.shutdown()
            self.httpd = None
        device.Device.close(self)

    def __init__(self, name, data):
        device.Device.__init__(self, name, data)
        self.closed = False
        self.httpd = None

        self.object_data_point("command", subtype="event")
        self.set_data_point("command", [None, time.time(), None])

        self.numeric_data_point("battery", min=0, max=100, unit="%", writable=False)
        self.set_alarm(
            "LowBattery",
            datapoint="battery",
            expression="value < 20",
            priority="warning",
            release_condition="value > 35",
        )

        try:
            self.set_config_default("device.serial", "P0A070000007")
            self.set_config_default("device.uuid", str(uuid.uuid4()))
            self.set_config_default("device.bind", "0.0.0.0:" + str(find_free_port()))

            if not self.config["device.bind"].strip():
                raise RuntimeError("No address selected")

            p = self.config["device.bind"].split(":")[1]
            p = int(p)

            self.bind = self.config["device.bind"]
            if not check_port(p):
                self.bind = "0.0.0.0:" + str(find_free_port())

            self.ssdp = HTTPUServer()
            self.ssdp.services = {
                "roku:ecp": {
                    "Location": "http://" + self.bind.replace("0.0.0.0", "localhost"),
                    "LOCATION": "http://" + self.bind.replace("0.0.0.0", "localhost"),
                    "USN": "uuid:roku:ecp:"
                    + self.config["device.uuid"].replace("-", "").upper(),
                    "Cache-Control": "max-age=3600",
                }
            }

            class S(BaseHTTPRequestHandler):
                def _set_headers(s):
                    s.send_response(200)
                    s.send_header("Content-type", "text/xml")
                    s.end_headers()

                def do_GET(s):
                    if not s.path.endswith(".png"):
                        s.send_response(200)
                        s.send_header("Content-Type", "text/xml; charset=utf-8")
                        s.end_headers()
                    else:
                        s.send_response(200)
                        s.send_header("Content-type", "image/png")
                        s.end_headers()

                    if s.path == "/query/apps":
                        x = b"""<apps>
                        <app id="11">Roku Channel Store</app>
                        <app id="12">Netflix</app>
                        <app id="13">Amazon Video on Demand</app>
                        <app id="837">YouTube</app>
                        <app id="2016">Crackle</app>
                        <app id="3423">Rdio</app>
                        <app id="21952">Blockbuster</app>
                        <app id="31012">MGO</app>
                        <app id="43594">CinemaNow</app>
                        <app id="46041">Sling TV</app>
                        <app id="50025">GooglePlay</app>
                        </apps>"""
                        x.wfile.write(x.encode())

                    if s.path == "/query/device-info":
                        s.wfile.write(fakeroku(self.name).encode())

                    elif s.path == "/":
                        s.wfile.write(
                            ssdpxml(self.name, self.config["device.uuid"]).encode()
                        )

                    elif s.path == "/unixtime":
                        s.wfile.write(str(time.time()).encode())

                    elif s.path == "/device-image.png":
                        with open(
                            os.path.join(os.path.dirname(__file__), "placeholder.png"),
                            "rb",
                        ) as f:
                            s.wfile.write(f.read())

                def do_HEAD(s):
                    s._set_headers()

                def do_POST(s):
                    s._set_headers()
                    s.wfile.write(b"{}")

                    if s.path.startswith("/keypress/"):
                        self.set_data_point(
                            "command",
                            [s.path[len("/keypress/") :], time.time(), None],
                        )
                    if s.path.startswith("/launch/"):
                        self.set_data_point(
                            "command",
                            [
                                "launch:" + s.path[len("/launch/") :],
                                time.time(),
                                None,
                            ],
                        )

            def f():
                try:
                    ip, port = self.bind.split(":")
                    self.httpd = HTTPServer((ip, int(port)), S)
                    self.httpd.serve_forever()
                except Exception:
                    self.handle_exception()

            self.thread = threading.Thread(
                target=f, name="HTTPRemoteServer" + self.name, daemon=True
            )
            self.thread.start()

            self.thread2 = threading.Thread(
                target=self.ssdploop, name="pyremotessdp" + self.name, daemon=True
            )
            self.thread2.start()

        except Exception:
            self.handleException()



from iot_devices import device
import gpiozero
from gpiozero.pins.mock import MockFactory


def str_to_bool(s: str):
    s = s.lower()
    if s in ('1', 'yes', 'true', 'on', 'enabled'):
        return True
    elif s in ('0', 'no', 'false', 'off', 'disabled'):
        return False
    else:
        raise ValueError("Not a valid boolean string")


class GPIOOutput(device.Device):
    device_type = "GPIOOutput"

    def __init__(self, name, data, **kw):
        device.Device.__init__(self, name, data, **kw)

        self.set_config_default("device.pin", "MOCK1")
        self.config_properties['device.active_high'] = {
            'description': 'This pin can be a mock pin like MOCK1, or any pin name supported by gpiozero, like BCM17 on a RasPi'
        }

        self.set_config_default("device.active_high", "1")
        self.config_properties['device.active_high'] = {
            'type': 'bool'
        }

        self.set_config_default("device.pwm_frequency", "100")

        self.set_config_default("device.initial_value", "0")
        self.config_properties['device.initial_value'] = {
            'type': 'float',
            'min': 0,
            'max': 1
        }
        try:
            driver = None
            pin = self.config['device.pin']

            freq = int(self.config['device.pwm_frequency'])

            if 'mock' in self.config['device.pin'].lower():
                driver = MockFactory()
                pin = pin.lower().replace('mock', '')

            # Push type data point set by the device
            self.numeric_data_point("value", writable=True, subtype='boolean', min=0, max=1, handler=self._set_pin,
                                    default=float(self.config['device.initial_value']))

            active_high = str_to_bool(self.config['device.active_high'])

            if pin:
                try:
                    self.pin = gpiozero.PWMLED(pin,
                                               active_high=active_high,
                                               # Host may have altered this default
                                               initial_value=self.datapoints['value'],
                                               pin_factory=driver,
                                               frequency=freq)
                except gpiozero.exc.PinPWMUnsupported:
                    self.print("Not support PWN on this pin")
                    self.pin = gpiozero.LED(pin,
                                            active_high=active_high,
                                            # Host may have altered this default
                                            initial_value=self.datapoints['value'],
                                            pin_factory=driver)

                self.metadata['driver'] = str(self.pin.pin_factory)
            else:
                self.pin = None
        except Exception:
            self.handle_exception()

    def close(self):
        if self.pin:
            self.pin.close()

    def _set_pin(self, v, t, a):
        if self.pin:
            self.pin.value = v

    @classmethod
    def discover_devices(cls, config={}, current_device=None, intent=None, **kw):
        return {}


class GPIOInput(device.Device):
    device_type = "GPIOInput"

    def __init__(self, name, data, **kw):
        device.Device.__init__(self, name, data, **kw)

        self.set_config_default("device.pin", "MOCK1")
        self.config_properties['device.active_high'] = {
            'description': 'This pin can be a mock pin like MOCK1, or any pin name supported by gpiozero, like BCM17 on a RasPi'
        }

        self.set_config_default("device.active_high", "true")
        self.config_properties['device.active_high'] = {
            'type': 'bool'
        }

        self.set_config_default("device.pull_up", "false")
        self.config_properties['device.pull_up'] = {
            'type': 'bool'
        }

        self.set_config_default("device.pull_down", "false")
        self.config_properties['device.pull_down'] = {
            'type': 'bool'
        }

        self.set_config_default("device.debounce_time_ms", "0")
        try:
            driver = None
            pin = self.config['device.pin']

            debounce = int(self.config['device.debounce_time_ms']) or None

            if 'mock' in self.config['device.pin'].lower():
                driver = MockFactory()
                pin = pin.lower().replace('mock', '')

            active_high = str_to_bool(self.config['device.active_high'])

            pull = None

            if str_to_bool(self.config['device.pull_up']):
                if str_to_bool(self.config['device.pull_down']):
                    raise ValueError("Can't have pull up and pull down")
                pull = True

                if active_high:
                    raise ValueError("Can't have pull up and active high")
                else:
                    active_high = None

            if str_to_bool(self.config['device.pull_down']):
                pull = False

                if not active_high:
                    raise ValueError("Can't have pull down and active low")
                else:
                    active_high = None

            if pin:
                self.pin = gpiozero.Button(pin,
                                           pull_up=pull,
                                           # Host may have altered this default
                                           active_state=active_high,
                                           pin_factory=driver,
                                           bounce_time=debounce)
                v = 1 if self.pin.value else 0

                self.pin.when_pressed = self.pressed
                self.pin.when_released = self.released

                self.metadata['driver'] = str(self.pin.pin_factory)
                # TODO race condition betweenset value and setup callbacks?
            else:
                v = 0
                self.pin = None

            self.numeric_data_point("value", writable=False, min=0, max=1,
                                    subtype='boolean',
                                    default=v)
            self.set_data_point('value', v)
        except Exception:
            self.handle_exception()

    def pressed(self):
        self.set_data_point('value', 1)

    def released(self):
        self.set_data_point('value', 0)

    def close(self):
        if self.pin:
            self.pin.close()



#!/usr/bin/python3

"""
# Run tui-dash, with the tui-dash.conf  as the first command line argument
# Or put your config in ~/tui-dash/tui-dash.conf

"""

import time
import sys
import configparser
import os
import threading
import logging
import urwid
import urwid.numedit
import traceback
import copy
import json

from iot_devices.host import get_class, discover
from iot_devices.device import Device

from typing import Dict


def main():
    user_conf_dir = os.path.expanduser("~/.config/tui-dash")
    os.makedirs(user_conf_dir, exist_ok=True)

    DEFAULT_CONF = """
[A demo device]
type=DemoDevice
"""

    try:
        file = sys.argv[1]
    except Exception:
        file = os.path.join(user_conf_dir, "tui-dash.conf")

        if not os.path.exists(file):
            with open(file, "w") as f:
                f.write(DEFAULT_CONF)

    with open(file) as f:
        cfg = configparser.ConfigParser()
        cfg.read_file(f)

    all_device_data = {}
    ordering_list = []

    for i in cfg.sections():
        data = cfg[i]
        data["name"] = i

        ordering_list.append(i)

        # Make it a real dict
        d = {}
        for j in data:
            d[j] = data[j]
        data = d

        all_device_data[i] = data

    class LineBox2(urwid.LineBox):
        def selectable(self):
            return True

    # Reverse lock? We release it from in the loop so other threads can act like they are part of the loop.

    lock = threading.RLock()
    # cols = urwid.Columns([])

    cols = urwid.GridFlow([], cell_width=60, h_sep=2, v_sep=1, align="left")
    loop = urwid.MainLoop(urwid.Filler(cols, "top"))

    def work(*a):
        # Let someone else do stuff
        lock.release()
        time.sleep(0.0001)

        lock.acquire()
        loop.draw_screen()
        loop.set_alarm_in(0.1, work)

    loop.set_alarm_in(0.1, work)

    class CustomButton(urwid.Button):
        button_left = urwid.Text("[", align="right")
        button_right = urwid.Text("]", align="left")

    def Button(*args, **kwargs):
        b = CustomButton(*args, **kwargs)
        b = urwid.AttrMap(b, "", "highlight")
        b = urwid.Padding(b, left=1, right=1)
        return b

    # Not assigned a display location yet
    pending_devices = {}

    loaded = False

    def customize(c):
        class Mixin:
            def __init__(self, name, *a, **k) -> None:
                pending_devices[name] = self
                self.txts = {}
                self.edits = {}

                with lock:
                    title = urwid.Text(("bold", self.title), "center", "any")
                    self.pile = urwid.Pile([title])

            def create_subdevice(self, cls, name: str, config: Dict, *a, **k):
                """
                Allows a device to create it's own subdevices.
                """

                fn = self.name + "." + name
                # Mix in the config for the data
                try:
                    if fn in all_device_data:
                        config.update(all_device_data[fn])
                except KeyError:
                    logging.exception(
                        "Probably a race condition. Can probably ignore this one."
                    )

                # Customize the class we are given

                m = Device.create_subdevice(self, customize(cls), name, config, *a, **k)

                if loaded:
                    if not self.config.get("hidden", False):
                        with lock:
                            cols.contents.append((self.pile, cols.options()))

                return m

            def on_data_change(self, point, value, timestamp, annotation):
                if isinstance(value, (int, float)):
                    value = round(value, 6)

                with lock:
                    if point in self.txts:
                        self.txts[point].set_text(str(value))
                    if point in self.edits:
                        self.edits[point].set_edit_text(str(value))

            def numeric_data_point(
                self, name, writable=True, default=0, unit="", subtype="", **kwargs
            ):
                c.numeric_data_point(self, name, **kwargs)

                with lock:
                    t = urwid.Text(("bold", name), "left", "any")

                    if not writable:
                        t2 = urwid.Text("no data ")
                        ut = urwid.Text(unit)

                        def refresh(*a):
                            self.request_data_point(name)

                        b2 = Button("get", refresh)

                        cols = urwid.Columns(
                            [(20, t), (8, t2), (6 + 8 + 2, ut), (9, b2)],
                            min_width=4,
                            dividechars=1,
                        )
                        self.txts[name] = t2

                    elif subtype == "trigger":

                        def set(*a):
                            self.set_data_point(
                                name, (self.datapoints.get(name, 0) or 0) + 1
                            )

                        b = Button("go", set)

                        cols = urwid.Columns(
                            [(29 + 8 + 6 + 2 + 1, t), (9, b)],
                            min_width=4,
                            dividechars=1,
                        )

                    elif subtype == "bool":
                        t2 = urwid.Text("no data ")

                        def refresh(*a):
                            self.request_data_point(name)

                        b2 = Button("get", refresh)

                        def toggle(*a):
                            self.set_data_point(
                                name,
                                0 if (self.datapoints.get(name, 0) or 0) > 0.5 else 1,
                            )

                        b2 = Button("Toggle", toggle)

                        cols = urwid.Columns(
                            [(20, t), (22, t2), (12, b2)], min_width=4, dividechars=1
                        )
                        self.txts[name] = t2

                    else:
                        t2 = urwid.numedit.FloatEdit("", str(default))
                        ut = urwid.Text(unit)

                        self.edits[name] = t2

                        def set(*a):
                            self.set_data_point(name, t2.value())

                        b = Button("set", set)

                        def refresh(*a):
                            self.request_data_point(name)

                        b2 = Button("get", refresh)

                        cols = urwid.Columns(
                            [(20, t), (8, t2), (6, ut), (9, b), (9, b2)],
                            min_width=4,
                            dividechars=1,
                        )
                        cols.set_focus(t2)

                    self.pile.contents.append((cols, ("pack", 1)))
                    self.pile.set_focus(cols)

                    # self.pile.contents.append(   (t2,('pack', 1)) )

            def print(self, *a):
                pass

        class ImportedDevice(Mixin, c):
            def __init__(self, name, *a, **k) -> None:
                Device.__init__(self, name, *a, **k)
                Mixin.__init__(self, name, *a, **k)
                c.__init__(self, name, *a, **k)

        return ImportedDevice

    devs = []

    for i, data in all_device_data.items():
        if (data.get("is_subdevice", False) in (True, "true")) or "type" not in data:
            # Name placeholder
            devs.append(i)
            continue

        # Get the class that would be able to construct a matching device given the data
        c = get_class(data)
        Cls = customize(c)

        # Make an instance of that device
        device = Cls(i, data)
        devs.append(device)

    for i in ordering_list:
        if i in pending_devices:
            if not pending_devices[i].config.get("hidden", False):
                cols.contents.append((pending_devices[i].pile, cols.options()))
                del pending_devices[i]

    for i, val in pending_devices.items():
        if not val.config.get("hidden", False):
            cols.contents.append((val.pile, cols.options()))

    loaded = True
    lock.acquire()
    loop.run()


if __name__ == "__main__":
    main()



# Plugin that manages YeeLight devices.

import yeelight
from mako.lookup import TemplateLookup
import os
import mako
import time
import threading
import logging
import random
import weakref
import traceback

from yeelight import discover_bulbs

import iot_devices.device

import colorzero

logger = logging.Logger("plugins.yeelight")

templateGetter = TemplateLookup(os.path.dirname(__file__))


lookup = {}

lastRefreshed = 0

lock = threading.Lock()


def maybeRefresh(t=30):
    global lastRefreshed
    if lastRefreshed < time.time() - t:
        refresh()
    elif lastRefreshed < time.time() - 5:
        refresh(3)


def refresh(timeout=8):
    global lastRefreshed, lookup
    from yeelight import discover_bulbs

    lastRefreshed = time.time()

    d = discover_bulbs()
    l = {}

    for i in d:
        if "name" in i["capabilities"]:
            l[i["capabilities"]["name"]] = i
        l[i["ip"]] = i
    lookup = l


def isIp(x):
    x = x.strip().split(".")
    try:
        x = [int(i) for i in x]
        if len(x) == 4:
            return True
    except Exception:
        return False


def getDevice(locator, timeout=10, klass=None):
    """Since plugs can change name, you should't keep a reference
    to a plug for too long. Instead use this function.
    """
    global lookup
    if not isIp(locator):
        if locator in lookup:
            return lookup[locator]

        maybeRefresh()
        if locator in lookup:
            return lookup[locator]

    return klass(locator)


def makeFlusher(wr):
    def f():
        while 1:
            try:
                time.sleep(3)
                d = wr()
                if not d:
                    return
                if d.closed:
                    return
                d.flush()
            except Exception as e:
                print(traceback.format_exc())

    t = threading.Thread(daemon=True, name="YeelightFlusher", target=f)
    t.start()


class YeelightDevice(iot_devices.device.Device):
    def __init__(self, name, data, **kw):
        self.lock = threading.Lock()
        iot_devices.device.Device.__init__(self, name, data, **kw)

        self.numeric_data_point("rssi", writable=False)
        self.set_alarm("Low Signal", "rssi", "value < -90")

        self.rssiCacheTime = 0
        self.lastLoggedUnreachable = 0

    def getRawDevice(self):
        return getDevice(self.data.get("device.locator"), 3, self.kdClass)

    def rssi(self, cacheFor=120, timeout=3):
        "These bulbs don't have RSSI that i found, instead we just detect reachability or not"
        with self.lock:
            "Returns the current RSSI value of the device"
            if time.time() - self.rssiCacheTime < cacheFor:
                return self.datapoints["rssi"] or -75

            # Not ideal, but we really can't be retrying this too often.
            # if it's disconnected. Way too much slowdown
            self.rssiCacheTime = time.time()

            try:
                info = getDevice(
                    self.data.get("device.locator"), timeout, self.kdClass
                ).get_properties()

                self.set_data_point("rssi", -70)
            except Exception:
                self.set_data_point("rssi", -180)
                if self.lastLoggedUnreachable < time.time() - 30:
                    self.handle_error("Device was unreachable")
                    self.lastLoggedUnreachable = time.time()
                raise

            return self.datapoints["rssi"] or -75


class YeelightRGB(YeelightDevice):
    device_type = "YeelightRGB"
    kdClass = yeelight.Bulb
    descriptors = {
        "kaithem.device.hsv": 1,
        "kaithem.device.powerswitch": 1,
    }

    def flush(self):
        if not self.hasData:
            return

        # Rate limits will prevent getting rejected
        self.allowedOperations = min(
            self.allowedOperations + (time.time() - self.lastRecalcedAllowed), 40
        )
        self.lastRecalcedAllowed = time.time()

        # Drop frames randomly. This will look a lot better with flickery type effects.
        if (random.random() * 20 + 1) > self.allowedOperations:
            return

        self.allowedOperations -= 1
        self.hasData = False

        color = colorzero.Color(self.datapoints["color"] or "white")
        rgb = color.rgb

        # Very crappy color correction done by trial and error
        hsv = colorzero.Color.from_rgb(
            rgb[0], max(0, rgb[1] - rgb[0] * 0.1), max(0, rgb[2] - rgb[0] * 0.1)
        ).hsv

        duration = self.datapoints["fade"] or 0

        if hsv.v < 0.0001:
            self.wasOff = True
            self.setSwitch(0, False, duration)
        else:
            try:
                self.getRawDevice().set_hsv(
                    int(hsv.h * 359.9),
                    int(hsv.s * 100),
                    int(hsv.v * 100),
                    effect="smooth" if duration else "sudden",
                    duration=int(duration * 1000) if not self.wasOff else 0,
                )
            except Exception:
                # Assume turned off.  Turn on. Retry.
                self.setSwitch(0, True, duration)
                self.getRawDevice().set_hsv(
                    int(hsv.h * 359.9),
                    int(hsv.s * 100),
                    int(hsv.v * 100),
                    effect="smooth" if duration else "sudden",
                    duration=int(duration * 1000) if not self.wasOff else 0,
                )

            if self.wasOff:
                self.wasOff = False
                self.setSwitch(0, True, duration)

    def close(self):
        return super().close()
        self.closed = True

    def __init__(self, name, data):
        YeelightDevice.__init__(self, name, data)
        self.closed = False
        self.lastHueChange = time.time()

        self.allowedOperations = 60
        self.lastRecalcedAllowed = time.time()

        # Has color data to flush
        self.hasData = False

        def swhandle(v, t, a):
            try:
                self.setSwitch(0, v >= 0.5)
            except Exception:
                pass

        self.numeric_data_point(
            "switch", min=0, max=1, subtype="bool", interval=300, handler=swhandle
        )
        self.numeric_data_point("fade", min=0, max=10, subtype="light_fade_duration")

        self.huesat = -1
        self.lastVal = -1
        self.wasOff = True
        self.oldTransitionRate = -1

        def colorhandle(v, t, a):
            self.hasData = True
            self.flush()

        self.string_data_point("color", subtype="color", handler=colorhandle)

        makeFlusher(weakref.ref(self))

    def getSwitch(self, channel, state):
        if channel > 0:
            raise ValueError("Bulb has 1 master power channel only")
        return self.getRawDevice().is_on

    def setSwitch(self, channel, state, duration=1):
        logger.debug(
            "Setting smartplug "
            + self.data.get("device.locator")
            + "to state "
            + str(state)
        )
        self.set_data_point("fade", duration)
        self.set_data_point("switch", 1 if state else 0)

        duration = self.datapoints["fade"] or 0

        with self.lock:
            "Set the state of switch channel N"
            if channel > 0:
                raise ValueError("This is a 1 channel device")
            try:
                if state:
                    getDevice(self.data.get("device.locator"), 3, self.kdClass).turn_on(
                        effect="smooth", duration=int(duration * 1000)
                    )
                    self.wasOff = False
                else:
                    getDevice(
                        self.data.get("device.locator"), 3, self.kdClass
                    ).turn_off(effect="smooth", duration=int(duration * 1000))
                    self.wasOff = True
                self.oldTransitionRate = duration
            except Exception:
                if self.lastLoggedUnreachable < time.time() - 30:
                    self.handle_error("Device was unreachable")
                    self.lastLoggedUnreachable = time.time()
                self.set_data_point("rssi", -180)
                raise

            self.set_data_point("rssi", -75)

    def setHSV(self, channel, hue, sat, val, duration=1):
        if channel > 0:
            raise ValueError("Bulb has 1 color only")
        self.set_data_point("fade", duration)
        self.set_data_point("color", colorzero.Color.from_hsv(hue / 360, sat, val).html)

    @classmethod
    def discover_devices(cls, config={}, current_device=None, intent=None, **kw):
        global lookup
        maybeRefresh()
        l = {}
        for i in lookup:
            config2 = config.copy()

            config2.update(
                {
                    "type": cls.device_type,
                    "device.locator": lookup[i]["capabilities"].get("name", "")
                    or lookup[i]["ip"],
                }
            )

            l[lookup[i]["capabilities"].get("name", "") or lookup[i]["ip"]] = config2

        return l



from __future__ import annotations
from typing import Any
import json
import time
from threading import Thread, RLock, Event
import niquests
import logging

from websockets.sync.client import connect
from websockets.exceptions import ConcurrencyError
from scullery import scheduling

import iot_devices.device
import iot_devices.util

from iot_devices.util import str_to_bool as s2b

ILLEGAL_NAME_CHARS = "{}|\\<>,?-=+)(*&^%$#@!~`\n\r\t\0"

logger = logging.getLogger(__name__)


# Maintains an auto-reconnecting websocket client
class ArduinoCogsClient(iot_devices.device.Device):
    device_type = "ArduinoCogsClient"

    def checker(self):
        with self.lock:
            h = self.thread_handle
            if h:
                if h.is_alive():
                    return

            if not self.should_run:
                return
            if self.last_start_time > (time.time() - 30):
                return

            self.last_start_time = time.time()

            t = Thread(
                target=self.thread, name="ArduinoCogsClient:" + self.name, daemon=True
            )
            t.start()

            self.thread_handle = t

    def handle_trouble_code_state(self, msg: str, val: float):
        level = "warning"
        if msg.startswith("E"):
            level = "error"
        elif msg.startswith("W"):
            level = "warning"
        elif msg.startswith("I"):
            level = "info"
        elif msg.startswith("D"):
            level = "debug"
        elif msg.startswith("C"):
            level = "critical"

        msg = msg.lower()
        if msg not in self.datapoints:
            self.numeric_data_point(
                msg, min=0, max=1, default=0, subtype="bool", writable=False
            )
            self.set_alarm(msg, msg, "value > 0", priority=level)

        if val > 0.0001:
            self.set_data_point(msg, 1, annotation="FromRemoteDevice")
        else:
            self.set_data_point(msg, 0, annotation="FromRemoteDevice")

    def handle_message(self, msg: str):
        d = json.loads(msg)
        if "__troublecodes__" in d:
            for i in d["__troublecodes__"]:
                self.handle_trouble_code_state(i, d["__troublecodes__"][i])

        if "__error__" in d:
            self.handle_error(d["__error__"])

        if "__notification__" in d:
            self.print(d["__notification__"])

        if "__success__" in d:
            self.print(d["__success__"])

        for i in d:
            self.on_var_msg(i, d[i])

    def on_var_msg(self, msg: str, val: float):
        if msg in self.scale_factors:
            val = float(val) / self.scale_factors[msg]
        if msg in self.ext_to_internal_names:
            msg = self.ext_to_internal_names[msg]
            self.set_data_point(msg, val, annotation="FromRemoteDevice")

    def makeHandler(self, var: str):
        def handler(val: float, ts: float, a: Any):
            if a == "FromRemoteDevice":
                return

            self.update_remote_on_reconnect[var] = (val, ts)
            self.sendVar(var, val)

        return handler

    def sendVar(self, var: str, val: float):
        if var in self.scale_factors:
            val = float(val) * self.scale_factors[var]

        try:
            if self.ws:
                with self.lock:
                    self.ws.send(json.dumps({var: val}))
        except Exception:
            self.handle_exception()
            if self.should_run:
                self.set_data_point("api_connected", 0)

    def thread(self):
        url = self.url
        url = url.split("//")[-1]
        if url.endswith("/"):
            url = url[:-1]

        wsurl = "ws://" + url + "/api/ws"
        device_info_url = "http://" + url + "/api/cogs.info"
        info_url = "http://" + url + "/api/cogs.tags"
        details_url = "http://" + url + "/api/cogs.tag"
        trouble_code_url = "http://" + url + "/api/cogs.trouble-codes"

        send_on_connect = iot_devices.util.str_to_bool(
            self.config.get("update_on_reconnect", "true")
        )
        try:
            r = niquests.get(device_info_url, timeout=15)
            r.raise_for_status()

            with connect(wsurl) as ws:
                self.suppress_connect_error = False
                try:
                    tc = niquests.get(trouble_code_url, timeout=5)
                    tc.raise_for_status()
                    assert isinstance(tc.text, str)
                    td = json.loads(tc.text)
                    for i in td:
                        self.handle_trouble_code_state(i, td[i])
                except Exception:
                    self.handle_exception()

                r = niquests.get(info_url, timeout=5)
                if not self.should_run:
                    return
                r.raise_for_status()
                assert isinstance(r.text, str)
                tagdata = json.loads(r.text)["tags"]

                for i in tagdata:
                    if "$" in i:
                        continue
                    if i.startswith("_"):
                        continue

                    n = i
                    for c in ILLEGAL_NAME_CHARS:
                        n = n.replace(c, "")
                    n = n.lower()

                    val: float = tagdata[i]
                    details = niquests.get(details_url + "?tag=" + i, timeout=5)
                    if not self.should_run:
                        return
                    details.raise_for_status()

                    assert isinstance(details.text, str)
                    dt = json.loads(details.text)
                    unit = dt["unit"]
                    scale = dt["scale"] or 1

                    subtype = ""

                    if unit in ("bool", "boolean"):
                        subtype = "bool"
                        unit = ""
                    if unit in ("bang", "trigger"):
                        subtype = "trigger"
                        unit = ""

                    if i not in self.datapoints:
                        readonly = dt.get("readonly", False)

                        self.ext_to_internal_names[i] = n
                        self.internal_to_ext_names[n] = i

                        if readonly or subtype == "trigger":
                            self.exclude_update_on_reconnect[i] = True

                        self.scale_factors[i] = float(scale)

                        # Trigger values are special.
                        # We don't want to trigger on old stuff
                        if not subtype == "trigger":
                            default = val / scale
                        else:
                            default = 0

                        self.numeric_data_point(
                            n,
                            min=dt["min"] / scale,
                            max=dt["max"] / scale,
                            subtype=subtype,
                            unit=unit,
                            default=default,
                            writable=not readonly,
                            handler=self.makeHandler(i),
                        )

                        # We need to handle data that the host may tell us about,
                        # which might have existed before the device itself.
                        if not subtype == "trigger":
                            try:
                                if not readonly:
                                    if self.datapoint_timestamps.get(i, 0):
                                        # Uh oh race condition between data and timestamp
                                        # But I think it's irrelevant for now
                                        self.update_remote_on_reconnect[i] = (
                                            self.datapoints[i],
                                            self.datapoint_timestamps[i],
                                        )
                            except Exception:
                                logger.exception(
                                    "Failed to handle data that was set before the device connected"
                                )

                    if not subtype == "trigger":
                        if not send_on_connect or (
                            i not in self.update_remote_on_reconnect
                        ):
                            self.set_data_point(
                                n, val / scale, annotation="FromRemoteDevice"
                            )

                self.ws = ws
                if send_on_connect:
                    for i in self.update_remote_on_reconnect:
                        if i not in self.exclude_update_on_reconnect:
                            self.sendVar(i, self.update_remote_on_reconnect[i][0])

                self.set_data_point("api_connected", 1)
                self.running = True

                pong: Event | None = None

                if not self.should_run:
                    return
                original = self.should_run

                while self.should_run == original:
                    try:
                        msg = ws.recv(10)
                    except TimeoutError:
                        with self.lock:
                            if pong:
                                if not pong.is_set():
                                    raise Exception("Timeout waiting for pong")

                            # Fixed val to raise concurrency error
                            # if there's a timeout
                            try:
                                pong = ws.ping("keepalive")
                            except ConcurrencyError:
                                raise OSError("WS Disconnected")

                        continue

                    if msg and self.should_run:
                        if isinstance(msg, str):
                            try:
                                self.handle_message(msg)
                            except Exception:
                                self.handle_exception()

        except Exception:
            if not self.suppress_connect_error:
                self.suppress_connect_error = True
                self.handle_exception()
            if self.should_run:
                self.set_data_point("api_connected", 0)

        try:
            if not self:
                return
        except Exception:
            self.running = 0
            return

        self.running = 0
        self.ws = None
        if self.should_run:
            self.set_data_point("api_connected", 0)

    def __init__(self, name: str, data: dict[str, str], **kw: Any):
        super().__init__(name, data, **kw)

        try:
            self.thread_handle = None
            self.suppress_connect_error = False
            self.numeric_data_point(
                "api_connected", min=0, max=1, default=0, subtype="bool", writable=False
            )

            self.set_config_default("update_on_reconnect", "true")
            self.config_properties["update_on_reconnect"] = {
                "title": "Update on Reconnect",
                "type": "bool",
                "default": "true",
                "description": "Update remote datapoints on reconnect, treating this client as the source of truth.",
            }

            self.set_config_default("intermittent_availability", "false")

            self.config_properties["intermittent_availability"] = {
                "title": "Intermittent Availability",
                "type": "bool",
                "default": "false",
                "description": "If false, treat device being offline as an error",
            }

            if not s2b(self.config.get("intermittent_availability", "false")):
                self.set_alarm(
                    "Disconnected",
                    "api_connected",
                    "value < 1",
                    priority="warning",
                    auto_ack=True,
                )

            self.set_config_default("url", "example.local")

            self.ext_to_internal_names: dict[str, str] = {}
            self.internal_to_ext_names: dict[str, str] = {}

            self.update_remote_on_reconnect: dict[str, Any] = {}
            self.exclude_update_on_reconnect: dict[str, bool] = {}

            self.last_start_time = 0
            self.scale_factors: dict[str, float] = {}
            self.ws = None
            self.lock = RLock()
            self.running = False
            self.should_run = time.time()
            self.url: str = self.config["url"]
            self.checker()
            self.scheduled = scheduling.scheduler.every(self.checker, 5)

        except Exception:
            self.handle_exception()
            return

    def close(self):
        self.should_run = False

        try:
            self.scheduled.unregister()
        except Exception:
            pass

        try:
            if self.ws:
                self.ws.close()
        except Exception:
            pass

        try:
            for i in range(500):
                if self.thread_handle and self.thread_handle.is_alive():
                    time.sleep(0.001)
        except Exception:
            pass

        return super().close()



from __future__ import annotations
import traceback
from typing import Any
from collections.abc import Callable
import logging
import time
import json
import copy
import weakref
import threading

# example device_manifest.json file that should be in any module declaring devices. keys are device type names.


# submodule values are the
# submodule you would have to import to get access to that
# device type. blank is the root.
# easy to read manually!
# """
# "devices": {
#   "Device":{
#     "submodule": ""
#     "description": "Device base class
#   }
# }
# """

devices_list_lock = threading.RLock()

all_devices: dict[str, weakref.ref[Device]] = {}
_all_devices: dict[str, weakref.ref[Device]] = {}


def key_to_title(k: str) -> str:
    title = k
    if title.startswith("device."):
        title = title.split("device.", 1)[-1]
    title = title.replace("_", " ").title()
    return title


def properties_to_schema(properties: dict[str, dict[str, Any]]):
    """Converts an old style legacy config properties dict to a json schema"""
    schema: dict[str, Any] = {
        "type": "object",
        "properties": {},
        "__auto_generated_by_iot_devices": True,
    }
    for i in properties:
        if i in ("name", "type"):
            continue
        title = key_to_title(i)

        schema["properties"][i] = {
            "type": "string",
            "description": properties[i].get("description", ""),
            "title": title,
        }

        if "enum" in properties[i]:
            schema["properties"][i]["enum"] = properties[i]["enum"]
        if "secret" in properties[i]:
            schema["properties"][i]["format"] = "password"
        if properties[i].get("type", False) in ("bool", "boolean"):
            schema["properties"][i]["enum"] = ["true", "false"]

    return schema


# Gonna overwrite these functions insude functions...
minimum = min
maximum = max


class LegacyConfigProperties:
    """Represents the old style config properties dict,
    and converts anything you do
    """

    def __init__(self, device: Device, config: dict[str, Any]):
        self.device = device
        self.config = config

    def __getitem__(self, key: str):
        return self.config[key]

    def __setitem__(self, key: str, value: dict[str, Any]):
        self.config[key] = value

        if self.device.json_schema and not self.device.json_schema.get(
            "__auto_generated_by_iot_devices", False
        ):
            raise ValueError(
                "Cannot set config_properties after json_schema has already been set"
            )
        self.device.json_schema = properties_to_schema(self.config)


class Device:
    """represents exactly one "device".
    should not be used to represent an interface to a large collection, use
    one instance per device.


    Note that this is meant to be subclassed twice.
    Once by the actual driver, and again by the
    host application, to detemine how to handle calls
    made by the driver.
    """

    # this name must be the same as the name of the device itself
    device_type: str = "Device"

    # This represents either a long text readme or an absolute path beginning with / to such
    readme: str = ""

    # Can set this directly instead of using the old style strings-only config
    json_schema: dict[str, Any] = {}

    def __init__(
        self,
        name: str,
        config: dict[str, str],
        subdevice_config: dict[str, Any] | None = None,
        **kw: Any,
    ):  # pylint: disable=unused-argument
        """

        The base class __init__ does nothing if
        called a second time, to simplify the complex
        inheritance.

        Attributes:

            title:
                Taken from config['title'] if possible, otherwise it is the name.

            json_schema:
                The JSON schema for the device, or else an empty dict.
                If a device uses the legacy config style, this will
                be autogenerated for you.

                Does not have to include non-device specific fields
                like "name" or "type".  Use get_full_schema() to get a
                full schema suitable for directly generating a UI.

            config:
                The current configuration of the device


        Args:
            name: must be a special char free string.
                It may contain slashes, for compatibility with hosts using that for heirarchy

            config:
                Must contain a "type" field that matches the device type.
                The idea is that the config object has everything needed to make
                the device.

                Subdevice configuration must have is_subdevice: True in
                save files so the host does not try to create it by itself.


                All other fields must be optional, and a blank
                unconfigured device should be creatable.
                The device should set it's own missing fields for
                use as a template.

                Device specific options should start with "device."
                or otherwise be unique.

                Options starting with temp. are reserved for
                device specific things that should not actually
                be saved.

                Options ending with __ are all reserved!

        """

        # Due to complex inheritance patterns, this could be called more than once
        if not hasattr(self, "__initial_setup"):
            config = copy.deepcopy(config)

            if config.get("type", self.device_type) != self.device_type:
                # Special placeholder
                if self.device_type not in ("unsupported", "placeholder"):
                    raise ValueError(
                        "Configured type "
                        + config.get("type", self.device_type)
                        + " does not match this class type:"
                        + str((config["type"], self, type))
                    )

            if subdevice_config and not callable(subdevice_config):
                raise ValueError("subdevice_config must be callable")

            self._subdevice_config = subdevice_config

            # here is where we keep track of our list of
            # sub-devices for each device.
            # Sub-devices will always have a name like
            # ParentDevice.ChildDevice
            self.subdevices: dict[str, Device] = {}

            # allows us to show large amounts of data that
            # do not warrant a datapoint, as it is unlikely anyone
            # would want to show them in a main listing,
            # and nobody wants to see them clutter up anything
            # or slow down the system when they change.
            # Putting data here should have no side effects.
            self.metadata: dict[str, Any] = {}

            # Raise error on bad data.
            json.dumps(config)

            self.config: dict[str, Any] = config

            # This is a legacy interface for very simple things, setting
            # this or any key will autogenerate JSON schemas
            self._config_properties = LegacyConfigProperties(self, {})

            self.title: str = key_to_title(self.config.get("title", "").strip() or name)

            self.__datapointhandlers: dict[
                str,
                Callable[
                    [
                        Any,
                        float,
                        Any,
                    ],
                    None,
                ],
            ] = {}
            self.datapoints: dict[
                str, int | float | str | bytes | dict[str, Any] | None
            ] = {}

            # Used mostly to determine if the data is still the default.
            self.datapoint_timestamps: dict[str, float] = {}

            # Functions that can be called to explicitly request a data point
            # That return the new value
            self.__datapoint_getters: dict[str, Callable[[], Any]] = {}

            self.name = name

            self.text_config_files: list[str] = []
            """
                Expose files in the config dir for easy editing if the framework supports it.
            """

            # hasattr checked later
            self.__initial_setup = True  # pylint: disable=unused-private-member

            with devices_list_lock:
                global all_devices
                _all_devices[name] = weakref.ref(self)
                all_devices = copy.deepcopy(_all_devices)

    @property
    def config_properties(self) -> dict[str, Any]:
        """Included for basic compatibility, hosts can stll work without the data but should
        upgrade to JSON schemas.
        """
        return self._config_properties

    @config_properties.setter
    def config_properties(self, value: dict[str, Any]):
        """Setter for devices using old style strings-only config files"""
        self._config_properties = LegacyConfigProperties(self, {})
        for i in value:
            self._config_properties[i] = value[i]

    def get_full_schema(self) -> dict[str, Any]:
        """Returns a full schema of the device. Including
        auto-generated properties, and generic things all devices should have."""
        d = copy.deepcopy(self.json_schema)
        if "properties" not in d:
            d["properties"] = {}
        assert "properties" in d
        assert isinstance(d["properties"], dict)

        d["type"] = "object"
        d["additionalProperties"] = False

        for i in self.config:
            if i not in d["properties"]:
                v = self.config[i]
                title = key_to_title(i)
                if isinstance(v, bool):
                    d["properties"][i] = {"type": "boolean", "title": title}
                elif isinstance(v, (int, float)):
                    d["properties"][i] = {"type": "integer", "title": title}
                elif isinstance(v, str):
                    d["properties"][i] = {"type": "string", "title": title}
                elif isinstance(v, list):
                    d["properties"][i] = {"type": "array", "title": title}
                else:
                    d["properties"][i] = {"type": "object", "title": title}

        if "is_subdevice" in self.config and self.config["is_subdevice"]:
            d["properties"]["is_subdevice"] = {"type": "boolean", "const": True}
        else:
            d["properties"].pop("is_subdevice", None)  # type: ignore
        d["properties"]["name"] = {"type": "string", "title": "Name"}
        d["properties"]["type"] = {"type": "string", "title": "Type"}
        d["properties"]["title"] = {"type": "string", "title": "Title"}
        d["properties"]["description"] = {
            "type": "string",
            "format": "markdown",
            "title": "Description",
        }

        return d

    @property
    def _uses_old_style_config(self):
        return not self.json_schema or self.json_schema.get(
            "__auto_generated_by_iot_devices", False
        )

    def create_subdevice(
        self, cls: type[Device], name: str, config: dict[str, Any], *a: Any, **k: Any
    ) -> object:
        """
        Args:
            cls: The class used to make the device
            name: The base name of the device.  The full name will be parent.basename
            config: The config as would be passed to any other device

        Returns:
            The device object


        Allows a device to create it's own subdevices.

        The host implementation must take the class, make whatever subclass
        is needed based on it,
        Then instantiate it as if the other parameters were given straight to
        the device.

        When the device is closed, the host must clean up all subdevices
        before cleaning up the master device.
        The host will put the device object into the parent device's subdevice
          dict.

        The host will rename the device to reflect that it is a subdevice.
        It's full name will be parent.basename.

        The host will allow configuration of the device like any other device.
        It will override whatever config that you give this function
        with the user config.

        However the entry in self.subdevices will always be exactly as given
        to this function, referred to as the base name

        The host will add is_subdevice=True to the config dict.
        """

        fn = f"{self.name}.{name}"
        config = copy.deepcopy(config)
        config["name"] = fn
        config["is_subdevice"] = "true"
        config["type"] = cls.device_type
        config["is_subdevice"] = "true"

        k = copy.copy(k)

        sd = cls(fn, config, *a, **k)

        self.subdevices[name] = sd
        return sd

    def get_config_folder(self, create: bool = True):
        """
        Devices may, in some frameworks, have their own folder in which they can place additional
        configuration, allowing for advanced features that depend on user content.

        Returns:
            An absolute path
        """

        # Can still call with create false just to check
        if create:
            raise NotImplementedError(
                "Your framework probably doesn't support this device"
            )

    @staticmethod
    def discover_devices(
        config: dict[str, Any],  # pylint: disable=unused-argument
        current_device: object | None = None,  # pylint: disable=unused-argument
        intent: str = "",  # pylint: disable=unused-argument
        **kwargs: Any,  # pylint: disable=unused-argument
    ) -> dict[str, dict[str, Any]]:
        """
        Discover a set of suggested configs that could be used to build a new device.

        Not required to be implemented and may just return {}

        ***********************
        Discovered suggestions MUST NOT have any passwords or secrets
        if the suggestion would cause them to be tried somewhere
        other than what the user provided them for,
        unless the protocol does not actually reveal the secrets
        to the server.

        You do not want to autosuggest trying the same credentials
        at bad.com that the user gave for example.com.


        The suggested UI semantics for discover commands is
        "Add a similar device" and "Reconfigure this device".

        Reconfiguration should always be available as the user
        might always want to take an existing device object and
        swap out the actual physical device it connects to.

        Kwargs is reserved for further hints on what kinds of
        devices should be discovered.

        Args:
            config: You may pass a partial config, or a completed
                config to find other
                similar devices. The device should reuse as much
                of the given config as possible and logical,
                discarding anything that wouldn't  work with the
                selected device.

            current_device: May be set to the current version of a
                device, if it is being used in a UI along the lines of
                suggesting how to further set up a partly configured
                device, or suggesting ways to add another
                similar device.

            kwargs: is reserved for further hints on what kinds
                of devices should be discovered.


            intent: may be a hint as to what kind of config you are
                    looking for.
                If it is "new", that means the host wants to add
                another similar device.  If it is "replace",
                the host wants to keep the same config
                but point at a different physical device.

                If it is "configure",  the host wants to look
                for alternate configurations available for the
                same exact device.

                If it is "step", the user wants to refine
                the existing config.

        Returns:
            A dict of device data dicts that could be used
            to create a new device, indexed by a descriptive name.

        """

        return {}

    def set_config_option(self, key: str, value: Any):
        """sets a top-level key in self.config. used for subclassing as you may want
        to persist.

        __init__ will automatically set the state.  this is used by the device
        itself to set it's own persistent values at runtime, perhaps in response
        to a websocket message.

        __init__ will automatically set the state when passed the config dict,
        you don't have to do that part.

        this is used by the device itself to set it's own persistent values at
        runtime, perhaps in response to a websocket message.

        the host is responsible for subclassing this and actually saving the
        data somehow, should that feature be needed.

        The device may subclass this to respond to realtime config changes.

        Devices should not clean up or get rid of keys they do not understand,
        because the host application may use (suitably prefixed for uniqueness) extra
        keys.
        """

        if isinstance(value, str):
            value = value.strip()

        # Auto strip the values to clean them up
        self.config[key] = value

    def set_config_default(self, key: str, value: str):
        """sets an option in self.config if it does not exist or is blank.
        Calls into set_config_option, you should not need to subclass this.
        """

        if key not in self.config or not self.config[key].strip():
            self.set_config_option(key, value.strip())

        if self.json_schema:
            if "properties" in self.json_schema:
                if key not in self.json_schema["properties"]:
                    title = key_to_title(key)
                    self.json_schema["properties"][key] = {
                        "type": "string",
                        "default": value,
                        title: title,
                    }

    def wait_ready(self, timeout: float = 15):  # pylint: disable=unused-argument
        """Call this to block for up to timeout seconds for the device to be fully initialized.
        Use this in quick scripts with a devices that readies itself asynchronously
        """
        return

    def print(self, s: str, title: str = ""):
        """used by the device to print to the hosts live device message feed, if such a thing should happen to exist"""
        logging.info(f"{title}: {str(s)}")

    def handle_error(self, s: str, title: str = ""):
        """like print but specifically marked as error. may get special notification.  should not be used for brief network loss"""
        logging.error(f"{title}: {str(s)}")

    def handle_exception(self):
        "Helper function that just calls handle_error with a traceback."
        self.handle_error(traceback.format_exc())

    def handle_event(self, event: str, data: Any | None):
        "Handle arbitrary messages from the host"

    def numeric_data_point(
        self,
        name: str,
        *,
        min: float | None = None,
        max: float | None = None,
        hi: float | None = None,  # pylint: disable=unused-argument
        lo: float | None = None,  # pylint: disable=unused-argument
        default: float | None = None,
        description: str = "",  # pylint: disable=unused-argument
        unit: str = "",  # pylint: disable=unused-argument
        handler: Callable[[float, float, Any], Any] | None = None,
        interval: float = 0,  # pylint: disable=unused-argument
        subtype: str = "",  # pylint: disable=unused-argument
        writable: bool = True,  # pylint: disable=unused-argument
        dashboard: bool = True,  # pylint: disable=unused-argument
        **kwargs: Any,  # pylint: disable=unused-argument
    ):
        """Register a new numeric data point with the given properties.

        Handler will be called when it changes.
        self.datapoints[name] will start out with tha value of None

        The intent is that you can subclass this and have your own implementation of data points,
        such as exposing an MQTT api or whatever else.

        Most fields are just extra annotations to the host.

        Args:
            min: The min value the point can take on
            max: The max value the point can take on

            hi: A value the point can take on that would be
                considered excessive
            lo: A value the point can take on that would be
                considered excessively low

            description: Free text

            unit: A unit of measure, such as "degC" or "MPH"

            default: If unset default value is None,
                or may be framework defined. Default does not trigger handler.

            handler: A function taking the value,timestamp,
                and annotation on changes.

            interval :annotates the default data rate the point
                will produce, for use in setting default poll
                rates by the host, if the host wants to poll.
                It does not mean the host SHOULD poll this,
                it only suggest a rate to poll at if the host
                has an interest in this data.

            writable:  is purely for a host that might subclass
                this, to determine if it should allow writing to the point.

            subtype: A string further describing the data
                type of this value, as a hint to UI generation.

            dashboard: Whether to show this data point in overview displays.

        """

        if min is None:
            minval: float = -(10**24)
        else:
            minval = min

        if max is None:
            maxval: float = 10**24
        else:
            maxval = max

        self.datapoints[name] = default

        def on_change_attempt(v1: float | None, t: float, a: Any):
            if v1 is None:
                return
            if callable(v1):
                v1 = v1()

            if v1 is not None:
                v: float = v1
            else:
                return

            v = float(v)

            v = minimum(maxval, v)
            v = maximum(minval, v)

            t = t or time.time()

            if self.datapoints[name] == v:
                # It's still considered a change if the previous value
                # was the default.
                if self.datapoint_timestamps.get(name, 0):
                    return

            self.datapoints[name] = v

            # Handler used by the device
            if handler:
                handler(v, t, a)

            self.on_data_change(name, v, t, a)

        self.__datapointhandlers[name] = on_change_attempt

    def string_data_point(
        self,
        name: str,
        *,
        description: str = "",  # pylint: disable=unused-argument
        unit: str = "",  # pylint: disable=unused-argument
        handler: Callable[[str, float, Any], Any] | None = None,
        default: str | None = None,
        interval: float = 0,  # pylint: disable=unused-argument
        writable: bool = True,  # pylint: disable=unused-argument
        subtype: str = "",  # pylint: disable=unused-argument
        dashboard: bool = True,  # pylint: disable=unused-argument
        **kwargs: Any,  # pylint: disable=unused-argument
    ):
        """Register a new string data point with the given properties.

        Handler will be called when it changes.
        self.datapoints[name] will start out with tha value of None

        Interval annotates the default data rate the point will produce, for use in setting default poll
        rates by the host, if the host wants to poll.

        Most fields are just extra annotations to the host.


        Args:
            description: Free text

            default: If unset default value is None, or may be framework defined. Default does not trigger handler.

            handler: A function taking the value,timestamp, and annotation on changes.

            interval: annotates the default data rate the point will produce, for use in setting default poll
                rates by the host if the host wants to poll.

                It does not mean the host SHOULD poll this,
                it only suggest a rate to poll at if the host has an interest in this data.

            writable:  is purely for a host that might subclass this, to determine if it should allow writing to the point.

            subtype: A string further describing the data type of this value, as a hint to UI generation.

            dashboard: Whether to show this data point in overview displays.
        """

        self.datapoints[name] = default

        def on_change_attempt(v: str | None, t, a):
            "This function handles the change detection by itself"
            if v is None:
                return
            if callable(v):
                v = v()
            v = str(v)
            t = t or time.time()

            if self.datapoints[name] == v:
                # It's still considered a change if the previous value
                # was the default.
                if self.datapoint_timestamps.get(name, 0):
                    return

            self.datapoints[name] = v

            # Handler used by the device
            if handler:
                handler(v, t, a)

            self.on_data_change(name, v, t, a)

        self.__datapointhandlers[name] = on_change_attempt

    def object_data_point(
        self,
        name: str,
        *,
        description: str = "",  # pylint: disable=unused-argument
        unit: str = "",  # pylint: disable=unused-argument
        handler: Callable[[dict, float, Any], Any] | None = None,
        interval: float = 0,  # pylint: disable=unused-argument
        writable=True,  # pylint: disable=unused-argument
        subtype: str = "",  # pylint: disable=unused-argument
        dashboard=True,  # pylint: disable=unused-argument
        **kwargs,  # pylint: disable=unused-argument
    ):
        """Register a new object data point with the given properties.   Here "object"
        means a JSON-like object.

        Handler will be called when it changes.
        self.datapoints[name] will start out with tha value of None

        Interval annotates the default data rate the point will produce, for use in setting default poll
        rates by the host, if the host wants to poll.

        Most fields are just extra annotations to the host.

        Args:
            description: Free text

            handler: A function taking the value,timestamp, and annotation on changes

            interval :annotates the default data rate the point will produce, for use in setting default poll
                rates by the host, if the host wants to poll.  It does not mean the host SHOULD poll this,
                it only suggest a rate to poll at if the host has an interest in this data.

            writable:  is purely for a host that might subclass this, to determine if it should allow writing to the point.

            subtype: A string further describing the data type of this value, as a hint to UI generation.

            dashboard: Whether to show this data point in overview displays.
        """

        self.datapoints[name] = None

        def on_change_attempt(v1: dict[str, Any] | None, t, a):
            if v1 is None:
                return

            v: dict[str, Any] = v1

            if callable(v):
                v = v()

            # Validate
            json.dumps(v)

            # Mutability trouble
            v = copy.deepcopy(v)

            t = t or time.time()

            if self.datapoints[name] == v:
                # It's still considered a change if the previous value
                # was the default.
                if self.datapoint_timestamps.get(name, 0):
                    return

            self.datapoints[name] = v

            # Handler used by the device
            if handler:
                handler(v, t, a)

            self.on_data_change(name, v, t, a)

        self.__datapointhandlers[name] = on_change_attempt

    def bytestream_data_point(
        self,
        name: str,
        *,
        description: str = "",  # pylint: disable=unused-argument
        unit: str = "",  # pylint: disable=unused-argument
        handler: Callable[[bytes, float, Any], Any] | None = None,
        writable: bool = True,  # pylint: disable=unused-argument
        dashboard: bool = True,  # pylint: disable=unused-argument
        **kwargs: Any,  # pylint: disable=unused-argument
    ):
        """register a new bytestream data point with the
        given properties. handler will be called when it changes.
        only meant to be called from within __init__.

        Bytestream data points do not store data,
        they only push it through.

        Despite the name, buffers of bytes may not be broken up or combined, this is buffer oriented,

        """

        self.datapoints[name] = None

        def on_change_attempt(v: bytes | None, t, a):
            if not v:
                return
            t = t or time.time()
            self.datapoints[name] = v

            # Handler used by the device
            if handler:
                handler(v, t, a)

            self.on_data_change(name, v, t, a)

        self.__datapointhandlers[name] = on_change_attempt

    def push_bytes(self, name: str, value: bytes):
        """Same as set_data_point but for bytestream data"""
        self.set_data_point(name, value)

    def set_data_point(
        self,
        name: str,
        value: int | float | str | bytes,
        timestamp: float | None = None,
        annotation: Any | None = None,
    ):
        """
        Set a data point of the device. may be called by the device itself or by user code.

        This is the primary api and we try to funnel as much as absolutely possible into it.

        things like button presses that are not actually
        "data points" can be represented as things like
        (button_event_name, timestamp) tuples in object_tags.

        things like autodiscovered ui can be done just by
        adding more descriptive metadata to a data point.

        Args:
            name: The data point to set

            value: The literal value.
                Use set_data_point_getter for a
                callable which will return such.

            timestamp: if present is a time.time() time.

            annotation: is an arbitrary object meant to be
                compared for identity,
                for various uses, such as loop prevention
                when dealting with network sync, when you need
                to know where a value came from.


        This must be thread safe, but the change detection
        could glitch out and discard if you go from A to B
        and back to A again.

        When there is multiple writers you will want
        to either do your own lock or ensure that
        dyou use unique values,
        like with an event counter.

        """
        if timestamp is None:
            timestamp = time.time()
        self.datapoint_timestamps[name] = timestamp
        self.__datapointhandlers[name](value, timestamp, annotation)

    def set_data_point_getter(self, name: str, getter: Callable):
        """Set the Getter of a datapoint, making it into an
        on-request point.
        The callable may return either the new value,
        or None if it has no new data.
        """
        self.__datapoint_getters[name] = getter

    def on_data_change(self, name: str, value: Any, timestamp: float, annotation: Any):
        """Used for subclassing, this is how you watch for
        data changes"""

    def request_data_point(self, name: str) -> Any:
        """Rather than just passively read, actively
        request a data point's new value.

        Meant to be called by external host code.

        """
        if name in self.__datapoint_getters:
            x = self.__datapoint_getters[name]()
            if x is not None:
                timestamp = time.time()
                # there has been a change! Maybe!  call a handler
                self.__datapointhandlers[name](x, timestamp, "From getter")

                self.datapoint_timestamps[name] = timestamp
                self.datapoints[name] = x
                return x

        return self.datapoints[name]

    def set_alarm(
        self,
        name: str,
        datapoint: str,
        expression: str,
        priority: str = "info",
        trip_delay: float = 0,
        auto_ack: bool = False,
        release_condition: str | None = None,
        **kw: Any,
    ):
        """declare an alarm on a certain data point.
        means we should consider the data point to be in an
        alarm state whenever the expression is true.

        used by the device itself to tell the host what
        it considers to be an alarm condition.

        the expression must look like "value > 90", where
        the operator can be any of the common comparision
        operators.

        you may set the trip delay to require it to stay
        tripped for a certain time,
        polling during that time and resettig if it is not
        tripped.

        the alarm remains in the alarm state till the release
        condition is met, by default it's just when the trip
        condition is inactive.  at which point it will need
        to be acknowledged by the user.


        these alarms should be considered "presets" that
        the user can override if possible.
        by default this function could just be a no-op,
        it's here because of kaithem_automation's alarm support,
        but many applications may be better off with full manual
        alarming.

        in kaithem the expression is arbitrary,
        but for this lowest common denominator definition
        it's likely best to
        limit it to easily semantically parsible strings.

        """

    def close(self):
        "Release all resources and clean up"
        for i in list(self.subdevices.keys()):
            self.subdevices[i].close()
            del self.subdevices[i]

    def on_delete(self):
        """
        release all persistent resources, used by the host
        app to tell the user the device is being permanently
        deleted.
        may be used to delete any files automatically created.
        """

    def update_config(self, config: dict[str, Any]):
        "Update the config dynamically at runtime. May be subclassed by the device, not the host.  Uses set_config_option to notify the host."
        for i in config:
            if not self.config.get(i, None) == config[i]:
                self.set_config_option(i, config[i])

        self.title = self.config.get("title", "").strip() or self.name



TRUE_STRINGS = ("true", "yes", "on", "enable", "active", "enabled", "1")
FALSE_STRINGS = ("false", "no", "off", "disable", "inactive", "disabled", "0")


def str_to_bool(s: str) -> bool:
    "Return a bool value if the str is one of the valid bool strings"
    if s.lower() in TRUE_STRINGS:
        return True
    elif s.lower() in FALSE_STRINGS:
        return False
    else:
        raise ValueError("Not a valid boolean string")



from iot_devices import device

import subprocess
import platform
import time
import threading

try:
    import niquests as requests
except ImportError:
    import requests
import socket
import re

imported_time = time.time()


def ping_ok(sHost) -> bool:
    try:
        subprocess.check_output(
            "ping -{} 1 {}".format(
                "n" if platform.system().lower() == "windows" else "c", sHost
            ),
            shell=True,
        )
    except Exception:
        return False

    return True


class ServerMonitor(device.Device):
    device_type = "ServerMonitor"

    config_properties = {
        "device.target": {
            "description": "Hostname or URL to ping.  If an http:// url is used, will poll with HTTP as well as ping."
        },
        "device.expect_pattern": {
            "description": "With HTTP URLs, expects to find string matching this regex in the returned data"
        },
    }

    def __init__(self, name, data, **kw):
        device.Device.__init__(self, name, data, **kw)

        self.set_config_default("device.target", "")
        self.set_config_default("device.expect_pattern", "")
        self.set_config_default("device.check_interval", "300")

        # Push type data point set by the device
        self.numeric_data_point("status", subtype="bool", writable=False)
        self.set_alarm("External Server Down", "status", "value<1", priority="error")

        self.stop_flag = id(self)

        t = threading.Thread(
            target=self.work_loop,
            daemon=True,
            name="Monitor for " + self.config["device.target"],
        )
        t.start()

    def close(self):
        self.stop_flag = False
        return super().close()

    def work_loop(self):
        "This runs until the device is closed"
        val = self.stop_flag
        url = self.config["device.target"]
        if not url:
            return

        first = False

        if (time.time() - imported_time) < 15:
            time.sleep(min(15 - (time.time() - imported_time), 2))

        while self.stop_flag == val:
            reachable = 1

            try:
                host = (
                    url.split("://", 1)[-1]
                    .split(":")[0]
                    .split("/")[0]
                    .split("?")[0]
                    .split("@")[-1]
                )

                if not ping_ok(host):
                    reachable = 0
                    if (time.time() - imported_time) > 100:
                        self.print("Unreachable host")

                try:
                    self.metadata["IP Address"] = socket.gethostbyname(host)
                except Exception:
                    if (time.time() - imported_time) > 100:
                        self.handle_exception()

                if reachable:
                    if url.startswith(("http://", "https://")):
                        try:
                            r = requests.get(url, timeout=5)
                            r.raise_for_status()

                            if self.config["device.expect_pattern"]:
                                r = re.search(
                                    self.config["device.expect_pattern"], r.text
                                )

                            if not r:
                                raise ValueError(
                                    "Server response does not match regex pattern"
                                )

                        except Exception:
                            if (time.time() - imported_time) > 100:
                                self.handle_exception()
                            reachable = 0

                # When booting up, let everything get to a steady state first so we don't
                # get spurious alerts. Until the ready point we can set it true but not false.
                if reachable or ((time.time() - imported_time) > 300):
                    self.set_data_point("status", reachable)

            except Exception:
                self.handle_exception()

            if reachable or first:
                time.sleep(float(self.config["device.check_interval"]))
            else:
                # Retry faster at first, try to get good data as soon as possible
                first = True
                time.sleep(30)
